/*
Derived from Examples in 6.2, Network Security with OpenSSL by Pravir Chandra, Matt Messier, John Viega. 
Copyright  2002 O'Reilly & Associates, Inc.

This duplicates the functionality of:
openssl enc -bf-cbc

The format of the data is:
Salted__
salt[8]
data in blocks of 8.
It's very easy.

The default openssl enc -bf-cbc uses md5 hash to generate the key and iv. This 
is bad because md5 has a collision problem. Collisions can be generated 
*/

#include <openssl/evp.h>
#include <openssl/rand.h>
#include <stdio.h>
#include <string.h>

// use a 256-bit key like they say instead of a 128-bit key.
#define KEY_LENGTH EVP_MAX_KEY_LENGTH
// Header means: blowfish cbc, iv is the first 8 bytes, 
// salt is the next 8 bytes, and the password is sha1 hashed.
#define BF_HEADER "Salted__"
#ifdef _DEBUG

#define VERBOSE 1

#else

#define VERBOSE 0

#endif


typedef unsigned char BYTE;

void select_random_key(BYTE *key, int b)
{
    RAND_bytes(key, b);
}

void print_hex(BYTE *data, int b)
{
    int i;

    for (i = 0;  i < b - 1;  i++)
        fprintf(stderr, "%02hhX:", data[i]);
    fprintf(stderr, "%02hhX\n", data[b - 1]);
}

unsigned char *simple_digest(char *buf, unsigned int len, int *olen)
{
    const EVP_MD  *m;
    EVP_MD_CTX    ctx;
    unsigned char *ret;

    char alg[] = "sha1";

    OpenSSL_add_all_digests(  );
    if (!(m = EVP_get_digestbyname(alg)))
        return NULL;
    if (!(ret = (unsigned char *)malloc(EVP_MAX_MD_SIZE)))
        return NULL;
    EVP_DigestInit(&ctx, m);
    EVP_DigestUpdate(&ctx, buf, len);
    EVP_DigestFinal(&ctx, ret, olen);
    return ret;
}

// Generate a key from an sha1 hash of the password and salt.
// The longest key that this can generate is 160 bits.
// Normally it is prudent to use a 128-bit key since it's commonly used.
void generate_key_old(char *key, const char *password, const BYTE *salt, const BYTE *iv, int b)
{
    // salt_len is constant 8 bytes.
    int plen = strlen(password);
    char pass2[plen+8];
    memcpy(pass2, password, plen);
    memcpy(pass2+plen, salt, 8);
    
    int md_len;
    unsigned char *pass_dgst;
    pass_dgst = simple_digest(pass2, plen+8, &md_len);
    // It's an error condition for the key to be too short.
    if(md_len < b)
    {
        fprintf(stderr, "Error: message digest is not long enough for the key: %i < %i.\n", md_len, b);
    }
    memcpy(key, pass_dgst, b);
    free(pass_dgst);
}

void generate_key(BYTE *key, const char *password, const BYTE *salt, BYTE *iv, int b)
{
    const EVP_CIPHER *cipher = NULL;
    cipher = EVP_bf_cbc(); // Error: EVP_get_cipherbyname("bf-cbc");
    const EVP_MD *dgst = NULL;
    dgst = EVP_md5();
    // This is a pretty interesting system.
    // read openssl-xxx/doc/EVP_BytesToKey.pod for interesting info.
    EVP_BytesToKey(cipher, dgst, salt, (BYTE *)password, 
        strlen(password), 1, key, iv);
}

void select_random_iv(char *iv, int b)
{
    RAND_pseudo_bytes(iv, b);
}

int seed_prng(int bytes)
{
    // This slows you down if you use it very much.
    if (!RAND_load_file("/dev/random", bytes))
        return 0;
    return 1;
}

char *encrypt_data(EVP_CIPHER_CTX *ctx, char *data, int inl, int *rb)
{
    char *ret;
    int  i, tmp, ol;

    ol = 0;
    ret = (char *)malloc(inl + EVP_CIPHER_CTX_block_size(ctx));
    for (i = 0;  i < inl / 100;  i++)
    {
        EVP_EncryptUpdate(ctx, &ret[ol], &tmp, &data[ol], 100);
        ol += tmp;
    }
    if (inl % 100)
    {
        EVP_EncryptUpdate(ctx, &ret[ol], &tmp, &data[ol], inl%100);
        ol += tmp;
    }
    EVP_EncryptFinal(ctx, &ret[ol], &tmp);
    *rb = ol + tmp;
    return ret;
}

char *decrypt_data(EVP_CIPHER_CTX *ctx, char *ct, int inl)
{
    /* We're going to null-terminate the plaintext under the assumption it's
     * non-null terminated ASCII text.  The null can be ignored otherwise.
     */
    char *pt = (char *)malloc(inl + EVP_CIPHER_CTX_block_size(ctx) + 1);
    int  ol;

    EVP_DecryptUpdate(ctx, pt, &ol, ct, inl);
    if (!ol) /* there's no block to decrypt */
    {
        free(pt);
        return NULL;
    }
    pt[ol] = 0;
    return pt;
}

int main(int argc, char *argv[])
{
    EVP_CIPHER_CTX ctx;
    BYTE           key[KEY_LENGTH]; 
    BYTE           iv[EVP_MAX_IV_LENGTH];
    char           *enc_data;
    char           data[1024];
    char           password[] = "test123";
    BYTE           salt[PKCS5_SALT_LEN];
    int            enc_size;

    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }

    // TODO: Secure password retreival algorithm
    
#ifdef ENCRYPT    
    //select_random_key(key, EVP_MAX_KEY_LENGTH);
    select_random_iv(salt, 8);
    
    generate_key(key, password, salt, iv, KEY_LENGTH);
    if(VERBOSE) print_hex(key, KEY_LENGTH);
    if(key == 0)
    {
        fprintf(stderr, "Error: Key generation failed.\n");
        return 1;
    }
    select_random_iv(iv, EVP_MAX_IV_LENGTH);
#endif

    // Read binary data up to 1024 bytes from the stdin.
    int data_size = fread(data, 1, 1024, stdin);

#ifdef ENCRYPT    
    EVP_EncryptInit(&ctx, EVP_bf_cbc(  ), key, iv);
    enc_data = encrypt_data(&ctx, data, data_size, &enc_size);
    //printf("Ciphertext is %d bytes.\n", enc_size);
    printf(BF_HEADER);
    //fwrite(iv, 8, 1, stdout);
    fwrite(salt, 8, 1, stdout);
    //printf("|%d|", data_size);
    fwrite(enc_data, enc_size, 1, stdout);
#endif

#ifdef DECRYPT
    char *out;
    char  final[EVP_MAX_BLOCK_LENGTH];

#ifndef ENCRYPT
    // Take the encrypted data from the stdin.
    if(0 != strncmp(data, BF_HEADER, 8))
    {
        fprintf(stderr, "Error: Header not found.\n");
        return 1;
    }
    //memcpy(iv, data + 8, 8);
    memcpy(salt, data + 8, 8);
    generate_key(key, password, salt, iv, KEY_LENGTH);
    if(VERBOSE) print_hex(key, KEY_LENGTH);
    enc_data = data + 16;
    enc_size = data_size - 16;
    data_size = 0;
#endif
    EVP_DecryptInit(&ctx, EVP_bf_cbc(  ), key, iv);
    int j;
    for(j=0; j<enc_size; j+=8)
    {
        out = decrypt_data(&ctx, enc_data + j, 8);
        // The first one is always null, ignore it or fix it.
        // TODO: You should probably try to understand it as well.
        if(out != 0)
        {
                fwrite(out, 8, 1, stdout);
                free(out);
        }
    }
    if (!EVP_DecryptFinal(&ctx, final, &data_size))
    {
        fprintf(stderr, "Error: Decrypt Final, padding incorrect.\n");
        abort(  );
    }
    fwrite(final, data_size, 1, stdout);
#endif
    return 0;
}
